@@ -1,12 +1,12 @@
 from __future__ import annotations
 from typing import Callable
 from math import ceil
+from pathlib import Path
 
 import torch
 from torch import tensor, is_tensor, arange, randint
-from torch.nn import Module
+from torch.nn import Module, Parameter
 import torch.nn.functional as F
-from torch.func import functional_call, vmap
 
 from beartype import beartype
 from beartype.door import is_bearable
@@ -30,12 +30,12 @@ def exists(v):
 def default(v, d):
     return v if exists(v) else d
 
+def divisible_by(num, den):
+    return (num % den) == 0
+
 def normalize(t, eps = 1e-6):
     return F.layer_norm(t, t.shape[-1:], eps = eps)
 
-def round_up_mult(num, mult):
-    return ceil(num / mult) * mult
-
 # class
 
 class EvoStrategy(Module):
@@ -45,15 +45,18 @@ class EvoStrategy(Module):
         self,
         model: Module,
         *,
-        environment: Callable[[Module], float],  # the environment is simply a function that takes in the model and returns a fitness score
+        environment: Callable[[Module], float | int | Tensor],  # the environment is simply a function that takes in the model and returns a fitness score
         num_generations,
         noise_population_size = 30,
-        learning_rate = 1e-3, # todo - optimizer
-        noise_scale = 1e-3,   # the noise scaling during rollouts with environment, todo - figure out right value and make sure it can also be customized per parameter name through a dict
-        param_names_to_optimize: list[str] | None = None,
+        learning_rate = 1e-3,               # todo - optimizer
+        noise_scale = 1e-3,                 # the noise scaling during rollouts with environment, todo - figure out right value and make sure it can also be customized per parameter name through a dict
+        params_to_optimize: list[str] | list[Parameter] | None = None,
         fitness_to_weighted_factor: Callable[[Tensor], Tensor] = normalize,
+        checkpoint_every = None,            # saving every number of generations
+        checkpoint_path = './checkpoints',
         cpu = False,
         accelerate_kwargs: dict = dict(),
+        reject_generation_fitnesses_if: Callable[[Tensor], bool] | None = None
     ):
         super().__init__()
 
@@ -64,20 +67,29 @@ class EvoStrategy(Module):
 
         self.environment = environment
 
-        param_names = set(dict(model.named_parameters()).keys())
+        named_parameters_dict = dict(model.named_parameters())
+
+        param_to_name_index = {param: name for name, param in named_parameters_dict.items()}
+
+        param_names = set(named_parameters_dict.keys())
 
         # default to all parameters to optimize with evo strategy
 
-        param_names_to_optimize = default(param_names_to_optimize, param_names)
+        params_to_optimize = default(params_to_optimize, param_names)
+
+        # if given as list of Parameter, convert to names
+
+        if is_bearable(params_to_optimize, list[Parameter]):
+            params_to_optimize = [param_to_name_index[param] for param in set(params_to_optimize)]
 
         # validate
 
-        assert all([name in param_names for name in param_names_to_optimize])
-        assert len(param_names_to_optimize) > 0, 'nothing to optimize'
+        assert all([name in param_names for name in params_to_optimize])
+        assert len(params_to_optimize) > 0, 'nothing to optimize'
 
         # sort param names and store
 
-        param_names_list = list(param_names_to_optimize)
+        param_names_list = list(params_to_optimize)
         param_names_list.sort()
 
         self.param_names_to_optimize = param_names_list
@@ -96,6 +108,17 @@ class EvoStrategy(Module):
         self.noise_scale = noise_scale
         self.learning_rate = learning_rate
 
+        # rejecting the fitnesses for a certain generation if this function is true
+
+        self.reject_generation_fitnesses_if = reject_generation_fitnesses_if
+
+        # checkpointing
+
+        self.checkpoint_every = checkpoint_every
+
+        self.checkpoint_folder = Path(checkpoint_path)
+        self.checkpoint_folder.mkdir(exist_ok = True)
+
     @property
     def device(self):
         return self.accelerate.device
@@ -142,6 +165,15 @@ class EvoStrategy(Module):
 
             model.add_noise_(noise_config)
 
+    def checkpoint(self, filename = 'evolved.model'):
+
+        if self.accelerate.is_main_process:
+
+            filepath = self.checkpoint_folder / f'{filename}.pt'
+            torch.save(self.model.state_dict(), str(filepath))
+
+        self.accelerate.wait_for_everyone()
+
     @torch.inference_mode()
     def forward(
         self
@@ -165,20 +197,20 @@ class EvoStrategy(Module):
 
         # through many generations
 
-        for index in range(self.num_generations):
-            generation = index + 1
+        generation = 1
+
+        while generation <= self.num_generations:
 
             # predetermine the seeds for each population
             # each seed is then used as a seed for all the parameters
 
+            synced_seed = None
+
             if is_distributed:
                 seed_for_pop = randint(0, int(1e9), (), device = self.device)
                 synced_seed = self.accelerate.reduce(seed_for_pop, reduction = 'sum').item()
-                randint_for_seed_pop = with_seed(synced_seed)(randint)
-            else:
-                randint_for_seed_pop = randint
 
-            seeds_for_population = randint_for_seed_pop(0, MAX_SEED_VALUE, (pop_size_round_up,))
+            seeds_for_population = with_seed(synced_seed)(randint)(0, MAX_SEED_VALUE, (pop_size_round_up,))
 
             # divy up work across machine
 
@@ -194,7 +226,7 @@ class EvoStrategy(Module):
                     fitnesses.append(0)
                     continue
 
-                individual_param_seeds = with_seed(individual_seed)(torch.randint)(0, MAX_SEED_VALUE, (self.num_params,))
+                individual_param_seeds = with_seed(individual_seed)(randint)(0, MAX_SEED_VALUE, (self.num_params,))
 
                 noise_config = dict(zip(self.param_names_to_optimize, individual_param_seeds.tolist()))
                 noise_config = {param_name: (seed, self.noise_scale) for param_name, seed in noise_config.items()}
@@ -217,6 +249,12 @@ class EvoStrategy(Module):
             if is_distributed:
                 fitnesses = self.accelerate.gather(fitnesses)
 
+            # validate fitnesses
+
+            if exists(self.reject_generation_fitnesses_if) and self.reject_generation_fitnesses_if(fitnesses):
+                self.print(f'[{generation}] fitnesses rejected')
+                continue
+
             # pass fitnesses to evolve function
 
             self.evolve_(
@@ -228,6 +266,20 @@ class EvoStrategy(Module):
 
             self.print(f'[{generation}] average fitness: {fitnesses.mean():.3f} | fitness std: {fitnesses.std():.3f}')
 
+            # maybe checkpoint
+
+            if (
+                exists(self.checkpoint_every) and
+                divisible_by(generation, self.checkpoint_every)
+            ):
+                self.checkpoint(f'evolved.model.{generation}.pt')
+
+            # increment generation
+
+            generation += 1
+
         self.print('evolution complete')
 
+        self.checkpoint(f'evolved.model.final.pt')
+
         self.accelerate.end_training()
